import numpy as np

import iupitermag._core as _iu

from . import currentsheet, internal


def trace_field_to_planet(
    start_positions: np.ndarray,
    internal_field: str | internal.InternalField = "JRM33",
    currentsheet_field: str | currentsheet.CurrentSheetField = "CON2020",
):
    """
    Trace the magnetic field from a collection of points to the planet.

    Args:
        start_positions (np.ndarray): Array of shape (N, 3) where N is the number
            of points and indices [:,0], [:, 1], and [:, 2] represent the X, Y,
            and Z coordinates of each point in the IAU coordinate system.
        internal_field (str | internal.InternalField): The internal field to use (default="JRM33").
        currentsheet_field (str | currentsheet.CurrentSheetField): The current sheet field to use
            (default="CON2020").

    Returns:
        traces (list[np.ndarray]): List of traces where each trace is of shape (M, 3) where M is
            the number of coordinates of each trace.  Each point of the trace is described by its
            cartesian representation in the IAU coordinate system.  In total there should be N
            traces, one for each starting position.
    """
    if isinstance(internal_field, str):
        internal_field = internal.InternalField(internal_field)

    if isinstance(currentsheet_field, str):
        currentsheet_field = currentsheet.CurrentSheetField(currentsheet_field)

    return _iu.trace_field_to_planet(
        np.asarray(start_positions, dtype=float).reshape(-1, 3),
        internal_field._field,
        currentsheet_field._field,
    )



import iupitermag._core as _iu

from .field import Field


class InternalField(Field):
    def __init__(self, typefield="", g=None, h=None, degree=None):
        """
        Class for handling the planet's internal field.

        Args:
            typefield (str): Type of planet field. Allowed values are
                "JRM09", "JRM33", "Custom"
            g (np.array): Legendre coefficient array g[n ,m]
            h (np.array): Legendre coefficient array h[n, m]
            degree (int): Degree of field

            g, h, and degree do not need to be specified for named fields.

        Returns:
            InternalField class object
        """
        self._field = _iu.PyInternalField(typefield, g, h, degree)

    def get_coefficients(self):
        """
        Get the coeffiicients of the defined internal field.

        Returns:
            g (np.array): Legendre coefficient (g) in units of nT.
            h (np.array): Legendre coefficient (h) in units of nT.
        """
        return self._field.get_coefficients()



"""
Methods to calculate field quantities like the JRM magnetic field
model and the CON2020 current sheet model.
"""

import numpy as np


def calc_assoc_legendre_poly(theta, N):
    """
    Calculates Gaussian normalized Legendre polynomials
    Assumes that the input is of the form P_nm(cos(theta)) NOT P_nm(x)
    """
    sintheta = np.sin(theta)
    costheta = np.cos(theta)

    P = np.zeros((N + 1, N + 1))
    K = np.zeros((N + 1, N + 1))
    diffP = np.zeros((N + 1, N + 1))

    P[0, 0] = 2
    diffP[0, 0] = 0

    n = 1
    P[1, 1] = sintheta * P[n - 1, n - 1]
    diffP[1, 1] = sintheta * diffP[n - 1, n - 1] + costheta * P[n - 1, n - 1]
    P[1, 0] = costheta * P[0, 0]
    diffP[1, 0] = costheta * diffP[0, 0] - sintheta * P[0, 0]

    for n in range(2, N + 1):
        P[n, n] = sintheta * P[n - 1, n - 1]
        diffP[n, n] = sintheta * diffP[n - 1, n - 1] + costheta * P[n - 1, n - 1]

        for m in range(0, n):
            K[n, m] = ((n - 1) ** 2 - m**2) / ((2 * n - 1) * (2 * n - 3))
            P[n, m] = costheta * P[n - 1, m] - K[n, m] * P[n - 2, m]
            diffP[n, m] = (
                costheta * diffP[n - 1, m]
                - sintheta * P[n - 1, m]
                - K[n, m] * diffP[n - 2, m]
            )

    return P, diffP


def calc_internal_field(r_a, theta, phi, g, h, N):
    """
    Calculates Psi, Br, Btheta, Bphi based on the spherical harmonics
    representation.
    """
    Br = 0.0
    Btheta = 0.0
    Bphi = 0.0

    a_r = 1.0 / r_a

    P, diffP = calc_assoc_legendre_poly(theta, N)

    sintheta = np.sin(theta)
    sinphi = np.sin(phi)
    cosphi = np.cos(phi)
    inv_sintheta = 0.0
    if np.abs(sintheta) > 1.0e-3:
        inv_sintheta = 1.0 / sintheta

    for n in range(0, N + 1):
        m = 0
        sinphi_prev = 0.0
        cosphi_prev = 1.0
        # Psi += a * (a_r)**(n + 1) * P[n, m] * g[n, m]
        Br += (a_r) ** (n + 2) * (n + 1) * P[n, m] * g[n, m]
        Btheta -= (a_r) ** (n + 2) * diffP[n, m] * g[n, m]
        Bphi += inv_sintheta * (a_r) ** (n + 2) * P[n, m] * m * (-h[n, m])

        for m in range(1, n + 1):
            sinmphi = sinphi_prev * cosphi + cosphi_prev * sinphi
            cosmphi = cosphi_prev * cosphi - sinphi_prev * sinphi

            Br += (
                (a_r) ** (n + 2)
                * (n + 1)
                * P[n, m]
                * (g[n, m] * cosmphi + h[n, m] * sinmphi)
            )

            Btheta -= (
                (a_r) ** (n + 2) * diffP[n, m] * (g[n, m] * cosmphi + h[n, m] * sinmphi)
            )

            Bphi += (
                inv_sintheta
                * (a_r) ** (n + 2)
                * P[n, m]
                * m
                * (g[n, m] * sinmphi - h[n, m] * cosmphi)
            )

            sinphi_prev = sinmphi
            cosphi_prev = cosmphi

    return np.array([Br, Btheta, Bphi])


def read_coefficients_file(filename, degree=None, doNormalize=False):
    """Reads the coefficients and return after normalization."""

    N = degree if degree else 30
    g = np.zeros((N + 1, N + 1))
    h = np.zeros((N + 1, N + 1))

    with open(filename, "r") as f:
        f.readline()
        for n in range(1, N + 1):
            for m in range(0, n + 1):
                g[n, m] = float(f.readline().split()[1])
                # print('g', n, m, g[n,m])

            for m in range(1, n + 1):
                h[n, m] = float(f.readline().split()[1])
                # print('h', n, m, h[n,m])

    # Normalize the Gauss coefficients
    if doNormalize:
        S = np.zeros((N + 1, N + 1))
        S[0, 0] = 1
        for n in range(1, N + 1):
            S[n, 0] = S[n - 1, 0] * (2 * n - 1) / n
            S[n, 1] = S[n, 0] * np.sqrt(n * 2.0 / (n + 1))

            g[n, 0] = g[n, 0] * S[n, 0]
            h[n, 0] = h[n, 0] * S[n, 0]
            g[n, 1] = g[n, 1] * S[n, 1]
            h[n, 1] = h[n, 1] * S[n, 1]

            for m in range(2, n + 1):
                S[n, m] = S[n, m - 1] * np.sqrt((n - m + 1) / (n + m))
                g[n, m] = g[n, m] * S[n, m]
                h[n, m] = h[n, m] * S[n, m]

    return g, h, N



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.patches import Ellipse

import iupitermag as im

plt.style.use("dark_background")


def plot_jupiter(ax, **kwargs):
    jupiter = Ellipse((0.0, 0.0), 2.0, 2.0 * (1.0 - 1.0 / 15.4), **kwargs)
    ax.add_patch(jupiter)


if __name__ == "__main__":
    internal_field = im.InternalField("JRM33")
    cs_field = im.CurrentSheetField("CON2020")

    starting_positions = np.array(
        [
            [-10.0, 0.0, 0.0],
            [-15.0, 0.0, 0.0],
            [-20.0, 0.0, 0.0],
            [-25.0, 0.0, 0.0],
            [10.0, 0.0, 0.0],
            [15.0, 0.0, 0.0],
            [20.0, 0.0, 0.0],
            [25.0, 0.0, 0.0],
        ]
    )

    trace = im.trace_field_to_planet(starting_positions, internal_field, cs_field)

    fig, ax = plt.subplots(1, 1, figsize=(6, 3), dpi=200)
    for t in trace:
        ax.plot(t[:, 0], t[:, 2], lw=1.0)
    ax.axhline(0.0, color="0.5", lw=0.25)
    ax.axvline(0.0, color="0.5", lw=0.25)
    plot_jupiter(ax, color="0.8", zorder=3)
    ax.set_xlabel(r"X [R$_J$]")
    ax.set_ylabel(r"Z [R$_J$]")
    ax.set_aspect("equal")
    fig.savefig("images/traced_field_lines.png", facecolor="k", bbox_inches="tight")
    plt.show()
    plt.close(fig)



import iupitermag._core as _iu

from .field import Field


class CurrentSheetField(Field):
    def __init__(self, typefield, params={}, integration_type="analytic"):
        """
        Class for working with Jupiter current sheet models.

        Args:
            typefield (str): Type of current sheet field. Current available options are
                'CON2020' or 'Custom'. If using a named field, the `params` argument
                does not need to be used.

            params (dict): The parameters of the current sheet. Only necessary for 'Custom' field.

            integration_type (str): Type of integration used. Options are 'analytic' (default) or
                'integral' (not currently implemented).
        """
        self._field = _iu.PyCurrentSheetField(typefield, params, integration_type)

    def get_params(self):
        """
        Get the parameters of this current sheet field model.

        Returns:
            params (dict): Dictionary containing all parameters for this current sheet model.
        """
        return self._field.get_params()



import numpy as np

import iupitermag._core as _iu


class Field:
    _field: _iu.PyInternalField | _iu.PyCurrentSheetField

    def __init__(self):
        """
        Base class for all field types. Contains generic routines like
        calc_field, map_calc_field, trace_field, etc.
        """
        pass

    def calc_field(self, r, theta, phi):
        """
        Calculate the field strength at a location defined in a spherical
        planetocentric coordinate system.

        Args:
            r (float): Radius in planetary radii
            theta (float): Co-latitude in radians
            phi (float): Azimuth in radians

        Returns:
            field (np.array): Magnetic field components [Br, Btheta, Bphi] in a
                array.
        """
        return self._field.calc_field(r, theta, phi)

    def calc_field_xyz(self, x, y, z):
        """
        Calculate the field at a location defined by (x, y, z) in a Cartesian
        coordinate system.

        Args:
            x (float): The X coordinate.
            y (float): The Y coordinate.
            z (float): The Z coordinate.

        Returns:
            field (np.ndarray): An array containing the magnetic field vector in
                cartesian coordinates (Bx, By, Bz)
        """
        return self._field.calc_field_xyz(x, y, z)

    def map_calc_field(self, positions):
        """
        Calculates [Br, Btheta, Bphi] for a collection of N points.

        Args:
            positions (np.ndarray): Array of spherical coordinates of shape (N, 3),
                where the last index refers to (r, theta, phi).

        Returns:
            field_rtp (np.ndarray): Array of (Br, Btheta, Bphi) of shape (N, 3)
                for each point.
        """
        return self._field.map_calc_field(
            np.asarray(positions, dtype=float).reshape(-1, 3)
        )

    def parmap_calc_field(self, positions):
        """
        Calculates [Br, Btheta, Bphi] for a collection of N points using Rayon
        for parallelization.

        Args:
            positions (np.ndarray): Array of spherical coordinates of shape (N, 3),
                where the last index refers to (r, theta, phi).

        Returns:
            field_rtp (np.ndarray): Array of (Br, Btheta, Bphi) of shape (N, 3)
                for each point.
        """
        return self._field.parmap_calc_field(
            np.asarray(positions, dtype=float).reshape(-1, 3)
        )

    def map_calc_field_xyz(self, positions):
        """
        Calculates [Bx, By, Bz] for a collection of N points. Uses cartesian coordinates
        as input.

        Args:
            positions (np.ndarray): Array of spherical coordinates of shape (N, 3),
                where the last index refers to (x, y, z) coordinates.

        Returns:
            field_xyz (np.ndarray): Array of shape (N, 3) containing (Bx, By, Bz)
                for each point.
        """
        return self._field.map_calc_field_xyz(
            np.asarray(positions, dtype=float).reshape(-1, 3)
        )

    def parmap_calc_field_xyz(self, positions):
        """
        Calculates [Bx, By, Bz] for a collection of N points. Uses cartesian coordinates
        as input and Rayon for parallelization.

        Args:
            positions (np.ndarray): Array of spherical coordinates of shape (N, 3),
                where the last index refers to (x, y, z) coordinates.

        Returns:
            field_xyz (np.ndarray): Array of shape (N, 3) containing (Bx, By, Bz)
                for each point.
        """
        return self._field.parmap_calc_field_xyz(
            np.asarray(positions, dtype=float).reshape(-1, 3)
        )
